/*
 * mwave.c 
 * mwave device driver.
 *
 *  Written By: Mike Sullivan IBM Corporation
 *
 *  Copyright (C) 1999 IBM Corporation
 *
 * This program is free software; you can redistribute it and/or modify      
 * it under the terms of the GNU General Public License as published by      
 * the Free Software Foundation; either version 2 of the License, or         
 * (at your option) any later version.                                       
 *                                                                           
 * This program is distributed in the hope that it will be useful,           
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             
 * GNU General Public License for more details.                              
 *                                                                           
 * NO WARRANTY                                                               
 * THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR        
 * CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT      
 * LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,      
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is    
 * solely responsible for determining the appropriateness of using and       
 * distributing the Program and assumes all risks associated with its        
 * exercise of rights under this Agreement, including but not limited to     
 * the risks and costs of program errors, damage to or loss of data,         
 * programs or equipment, and unavailability or interruption of operations.  
 *                                                                           
 * DISCLAIMER OF LIABILITY                                                   
 * NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY   
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL        
 * DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND   
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    
 * USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED  
 * HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES             
 *                                                                           
 * You should have received a copy of the GNU General Public License         
 * along with this program; if not, write to the Free Software               
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *                                                                           
 * 
 *  10/23/2000 - Alpha Release 0.1.0
 *            First release to the public
 *
 */

/*
 * Ported from Linux to FreeBSD by Simon Walton
 * April 2001
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/ioccom.h>

#if 0
#include <isa/isavar.h>
#include <isa/pnpvar.h>

    /* PnP doesn't work */

static struct isa_pnp_id mwave_pnp[] =
    {
    { 0x80374d24, "IBM Mwave modem"},	/* <IBM3780> */
    {0, 0}
    };

#endif

#undef MW_TRACE

#include "smapi.h"
#include "mwavedd.h"
#include "tp3780i.h"

static int mwave_open(dev_t mdev, int oflags, int devtype, struct proc *p);
static int mwave_close(dev_t mdev, int fflag, int devtype, struct proc *p);
static int mwave_ioctl(dev_t mdev, u_long cmd, caddr_t data, 
    int fflag, struct proc *p);
static int mwave_read(dev_t dev, struct uio *uio, int ioflag);
static int mwave_write(dev_t dev, struct uio *uio, int ioflag);

static devclass_t mwave_devclass;

#define DEV2DEVICE(dev) devclass_get_device (mwave_devclass, minor (dev))


static struct cdevsw mwave_cdevsw =
    {
    mwave_open,			/* open */
    mwave_close,		/* release */
    mwave_read,			/* read */
    mwave_write,		/* write */
    mwave_ioctl,		/* ioctl */
    nopoll,			/* poll */
    nommap,			/* mmap */
    nostrategy,			/* strategy */
    "mwave",			/* name */
    MWAVE_MAJOR,		/* major */
    nodump,			/* dump */
    nopsize,			/* psize */
    0,				/* flags */
    -1				/* bmajor */
    };


static int mwave_open(dev_t mdev, int oflags, int devtype, struct proc *p)
{
  unsigned int retval=0;
  if (minor (mdev)) retval = ENODEV;

  PRINTK_2(TRACE_MWAVE,"mwave1::mwave_open, entry dev %x \n", mdev);
  PRINTK_2(TRACE_MWAVE,"mwave1::mwave_open, exit return retval %x\n",retval);

  return retval;
}

static int mwave_close(dev_t mdev, int fflag, int devtype, struct proc *p)
{
  unsigned int retval=0;
  
  PRINTK_2(TRACE_MWAVE,"mwave1::mwave_close, entry dev %x\n", mdev);
  
  PRINTK_2(TRACE_MWAVE,"mwave1::mwave_close, exit retval %x\n",retval);
    
  return retval;
}

static int mwave_ioctl(dev_t mdev, u_long cmd, caddr_t data, 
    int fflag, struct proc * current)
{
  pMWAVE_DEVICE_DATA pDrvData;
  unsigned int retval=0;

  pDrvData = device_get_softc (DEV2DEVICE (mdev));
  if (!pDrvData) return (ENXIO);

  switch(cmd) {

  case IOCTL_MW_RESET:
      PRINTK_1(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_RESET calling ResetDSP\n");
      retval=tp3780I_ResetDSP(&pDrvData->rBDData);
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_RESET retval %x from ResetDSP\n",retval);
      break;

  case IOCTL_MW_RUN:
      PRINTK_1(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_RUN calling StartDSP\n");
      retval=tp3780I_StartDSP(&pDrvData->rBDData);
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_RUN retval %x from StartDSP\n",retval);
      break;

  case IOCTL_MW_DSP_ABILITIES:
    {
      MW_ABILITIES rAbilities;
      
      PRINTK_1(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_DSP_ABILITIES calling QueryAbilities\n");
      retval=tp3780I_QueryAbilities(&pDrvData->rBDData,&rAbilities);
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_DSP_ABILITIES retval %x from QueryAbilities\n",retval);
      if (retval == 0) {
	  bcopy (&rAbilities, data, sizeof (MW_ABILITIES));
      }
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl, IOCTL_MW_DSP_ABILITIES exit retval %x\n",retval);
    }
    break;

  case IOCTL_MW_READ_DATA:
  case IOCTL_MW_READCLEAR_DATA:
    {
      MW_READWRITE * rReadData;
      unsigned short *pusBuffer=0;
      
      rReadData = (MW_READWRITE*) data;
      pusBuffer=(unsigned short *)(rReadData -> pBuf);
      
      PRINTK_4(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_READ_DATA, size %lx, ioarg %lx pusBuffer %p\n",rReadData -> ulDataLength,data,pusBuffer);
      retval=tp3780I_ReadWriteDspDStore(&pDrvData->rBDData,cmd,(void *)pusBuffer,rReadData -> ulDataLength,rReadData -> usDspAddress);
    }
    break;
    
  case IOCTL_MW_READ_INST:
    {
      MW_READWRITE * rReadData;
      unsigned short *pusBuffer=0;
      
      rReadData = (MW_READWRITE*) data;
      pusBuffer=(unsigned short *)(rReadData -> pBuf);
      
      PRINTK_4(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_READ_INST, size %lx, ioarg %lx pusBuffer %p\n",rReadData -> ulDataLength/2,data,pusBuffer);
      retval=tp3780I_ReadWriteDspDStore(&pDrvData->rBDData,cmd,pusBuffer,rReadData -> ulDataLength/2,rReadData -> usDspAddress);
    }
    break;
    
  case IOCTL_MW_WRITE_DATA:
    {
      MW_READWRITE * rWriteData;
      unsigned short *pusBuffer=0;
      
      rWriteData = (MW_READWRITE*) data;
      pusBuffer=(unsigned short *)(rWriteData -> pBuf);
      
      PRINTK_4(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_WRITE_DATA, size %lx, ioarg %lx pusBuffer %p\n",rWriteData -> ulDataLength,data,pusBuffer);
      retval=tp3780I_ReadWriteDspDStore(&pDrvData->rBDData,cmd,pusBuffer,rWriteData -> ulDataLength,rWriteData -> usDspAddress);
    } 
    break;
    
  case IOCTL_MW_WRITE_INST:
    {
      MW_READWRITE * rWriteData;
      unsigned short *pusBuffer=0;
      
      rWriteData = (MW_READWRITE*) data;
      pusBuffer=(unsigned short *)(rWriteData -> pBuf);
      
      PRINTK_4(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_WRITE_INST, size %lx, ioarg %lx pusBuffer %p\n",rWriteData -> ulDataLength,data,pusBuffer);
      retval=tp3780I_ReadWriteDspIStore(&pDrvData->rBDData,cmd,pusBuffer,rWriteData -> ulDataLength,rWriteData -> usDspAddress);
    }
    break;

  case IOCTL_MW_REGISTER_IPC:
    {
      unsigned int ipcnum=*(unsigned int*)data;
      
      if (ipcnum >= 16) {
	printf ("mwave::mwave_ioctl IOCTL_MW_REGISTER_IPC ERROR invalid ipcnum %x\n",ipcnum);
	return EINVAL;
      }
      PRINTK_3(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_REGISTER_IPC ipcnum %x entry usIntCount %x\n",ipcnum,
	     pDrvData->IPCs[ipcnum].usIntCount);

      pDrvData->IPCs[ipcnum].bIsHere=FALSE;
      pDrvData->IPCs[ipcnum].bIsEnabled=TRUE;
      current->p_nice=-16; /* boost to provide priority timing */
      
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_REGISTER_IPC ipcnum %x exit\n",ipcnum);
    }
    break;

  case IOCTL_MW_GET_IPC:
    {
      unsigned int ipcnum= *(unsigned int*)data;
      int s;
      
      if (ipcnum >= 16) {
	printf ("mwave::mwave_ioctl IOCTL_MW_GET_IPC ERROR invalid ipcnum %x\n",ipcnum);
 	return EINVAL;
      }
      PRINTK_3(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_GET_IPC ipcnum %x, usIntCount %x\n",ipcnum,pDrvData->IPCs[ipcnum].usIntCount);
      
      if (pDrvData->IPCs[ipcnum].bIsEnabled == TRUE) {
	PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl, thread for ipc %x going to sleep\n",ipcnum);

	s = spltty();
	/* check whether an event was signalled by the interrupt handler while we were gone */
	if (pDrvData->IPCs[ipcnum].usIntCount == 1) { /* first int has occurred (race condition) */
	  pDrvData->IPCs[ipcnum].usIntCount=2; /* first int has been handled */
	  PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_GET_IPC ipcnum %x handling first int\n",ipcnum);
	} else { /* either 1st int has not yet occurred, or we have already handled the first int */
	  pDrvData->IPCs[ipcnum].bIsHere=TRUE;
	  retval = tsleep (&pDrvData->IPCs[ipcnum].ipc_wait_queue, PCATCH|PSOCK,
	      "mwave", 0);
	  if (retval == 0)
	     {
	     pDrvData->IPCs[ipcnum].bIsHere=FALSE;
	     if (pDrvData->IPCs[ipcnum].usIntCount == 1) { 
	       pDrvData->IPCs[ipcnum].usIntCount=2; 
	       }
	      }
	  else if (retval == ERESTART) retval = EINTR;

	  splx (s);
	  PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_GET_IPC ipcnum %x woke up and returning to application\n",ipcnum);
	} 
	PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_GET_IPC, returning thread for ipc %x processing\n",ipcnum);
      }
    }
    break;
      
  case IOCTL_MW_UNREGISTER_IPC:
    {
      uint ipcnum= *(unsigned int*)data;
      
      PRINTK_2(TRACE_MWAVE,"mwave::mwave_ioctl IOCTL_MW_UNREGISTER_IPC ipcnum %x\n",ipcnum);
      if (ipcnum >= 16) {
	printf( "mwave::mwave_ioctl IOCTL_MW_UNREGISTER_IPC ERROR invalid ipcnum %x\n",ipcnum);
	return EINVAL;
      }
      if (pDrvData->IPCs[ipcnum].bIsEnabled == TRUE) {
	pDrvData->IPCs[ipcnum].bIsEnabled=FALSE;
	if (pDrvData->IPCs[ipcnum].bIsHere == TRUE) {
	  wakeup (&pDrvData->IPCs[ipcnum].ipc_wait_queue);
	}
      }
    }
    break;
    
  default:
    printf ("mwave1::mwave_ioctl, cmd %lux\n",cmd);
    return EINVAL;
    break;
  }
  
  return retval;
}


static int mwave_read(dev_t dev, struct uio *uio, int ioflag)
{
  PRINTK_1(TRACE_MWAVE,"mwave_read entry file\n");
  return 0;
}

static int mwave_write(dev_t dev, struct uio *uio, int ioflag)
{
  PRINTK_1(TRACE_MWAVE,"mwave_write entry\n");
  return 0;
}
  
static int mwave_attach (device_t dev)
{
  pMWAVE_DEVICE_DATA pDrvData;
  int retval=0,i;

  pDrvData = device_get_softc (dev);

  bzero (pDrvData, sizeof(MWAVE_DEVICE_DATA));
  pDrvData -> rBDData.mwave_dev = make_dev (&mwave_cdevsw, 0, UID_ROOT, 
      GID_WHEEL, 0600, "mwave");
  
  PRINTK_1(TRACE_MWAVE,"mwave::mwave_init entry\n");

  pDrvData->bBDInitialized=FALSE;
  pDrvData->bResourcesClaimed=FALSE;
  pDrvData->bDSPEnabled=FALSE;
  pDrvData->bDSPReset=FALSE;

  for(i=0;i<16;i++) {
    pDrvData->IPCs[i].bIsEnabled=FALSE;
    pDrvData->IPCs[i].bIsHere=FALSE;
    pDrvData->IPCs[i].usIntCount=0; /* no ints received yet */
  }
  
  retval=mwave_Initialize(&pDrvData->rBDData);
  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init, return from Initialize retval %x\n",retval);
  if (retval) {
    printf ("Mwave Modem, ERROR cannot Initialize DSP error %x\n",retval);
    goto cleanup;
  }

  pDrvData->bBDInitialized=TRUE;

  retval=mwave_CalcResources(&pDrvData->rBDData);
  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init, return from CalcResources retval %x\n",retval);
  if (retval) {
    printf ("Mwave Modem, ERROR cannot CalcResources error %x\n",retval);
    goto cleanup;
  }
  
  retval=mwave_ClaimResources(&pDrvData->rBDData);
  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init, return from ClaimResources retval %x\n",retval);
  if (retval) {
    printf ("Mwave Modem, ERROR cannot ClaimResources error %x\n",retval);
    goto cleanup;
  }  

  pDrvData->bResourcesClaimed=TRUE;

  retval=tp3780I_EnableDSP(dev, &pDrvData->rBDData);
  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init, return from tp3780I_EnableDSP retval %x\n",retval);
  if (retval) {
    printf ("Mwave Modem, ERROR cannot Enable DSP error %x\n",retval);
    goto cleanup;
  }    
  
  pDrvData->bResourcesClaimed=TRUE;

  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init exit retval %x\n",retval);
  return retval;
  
 cleanup:
  if (pDrvData->bDSPEnabled) {
    tp3780I_DisableDSP(dev, &pDrvData->rBDData);
  }
  if (pDrvData->bResourcesClaimed) {
    mwave_ReleaseResources(dev, &pDrvData->rBDData);
  }

  PRINTK_2(TRACE_MWAVE,"mwave::mwave_init exit retval %x\n",retval);
  return retval;
}

#if 0
/*
    /proc entry stuff added by Ian Pilcher <pilcher@us.ibm.com>
*/

#include <linux/proc_fs.h>

static int mwave_get_info(char *buf, char **start, off_t offset, int len);

static int mwave_get_info(char *buf, char **start, off_t offset, int len)
{
  DSP_3780I_CONFIG_SETTINGS *pSettings = &s_mdd.rBDData.rDspSettings;

  char *out = buf;

  out += sprintf (out,        "DSP_IRQ  %i\n", pSettings->usDspIrq);
  out += sprintf (out, "DSP_DMA  %i\n", pSettings->usDspDma);
  out += sprintf (out, "DSP_IO   %#.4x\n", pSettings->usDspBaseIO);
  out += sprintf (out, "UART_IRQ %i\n", pSettings->usUartIrq);
  out += sprintf (out, "UART_IO  %#.4x\n", pSettings->usUartBaseIO);
  return out - buf;
}


int mwave_read_proc (char *buf, char **start, off_t offset, int xlen, int
	unused)
{
    DSP_3780I_CONFIG_SETTINGS *pSettings = &s_mdd.rBDData.rDspSettings;
    int len;

    len = sprintf (buf,        "DSP_IRQ  %i\n", pSettings->usDspIrq);
    len += sprintf (&buf[len], "DSP_DMA  %i\n", pSettings->usDspDma);
    len += sprintf (&buf[len], "DSP_IO   %#.4x\n", pSettings->usDspBaseIO);
    len += sprintf (&buf[len], "UART_IRQ %i\n", pSettings->usUartIrq);
    len += sprintf (&buf[len], "UART_IO  %#.4x\n", pSettings->usUartBaseIO);
    return len;
}

static struct proc_dir_entry mwave_proc =
{
    0,					/* unsigned short low_ino */
    5,                          	/* unsigned short namelen */
    "mwave",                    	/* const char *name */
    S_IFREG | S_IRUGO,			/* mode_t mode */
    1,					/* nlink_t nlink */
    0,					/* uid_t uid */
    0,					/* gid_t gid */
    0,					/* unsigned long size */
    NULL,				/* struct inode_operations *ops */
    &mwave_read_proc			/* int (*get_info) (...) */
};

/*
    End of /proc entry stuff (except for the calls to proc_register and
    proc_unregister in init_module and cleanup_module)
*/
#endif

static void mwave_identify (driver_t *driver, device_t parent)
{
device_t dev;

dev = BUS_ADD_CHILD (parent, 0, "mwave", -1); 
device_set_driver (dev, driver);
if (bootverbose)
    device_printf (parent, "mwave: mwave_identify\n");
}

static int mwave_probe (device_t dev)
{
#if 0
int error;

error = ISA_PNP_PROBE(device_get_parent(dev), dev, mwave_pnp);
if (error == ENXIO) return error;
#endif

device_set_desc (dev, "mwave");
return 0;
}
 
static int mwave_detach (device_t dev)
{
pMWAVE_DEVICE_DATA pDrvData;

  pDrvData = device_get_softc (dev);

if (pDrvData->bDSPEnabled) {
    tp3780I_DisableDSP(dev, &pDrvData->rBDData);
    }

if (mwave_ReleaseResources(dev, &pDrvData->rBDData))
    PRINTK_ERROR ("mwave::mwave_detach, mwave_ReleaseResources failed\n");

destroy_dev (pDrvData -> rBDData.mwave_dev);
/* device_delete_child (device_get_parent (dev), dev); */
return 0;
}

unsigned short s_ausThinkpadIrqToField[16];
unsigned short s_ausThinkpadDmaToField[8];

static int mwave_suspend (device_t dev)
{
pMWAVE_DEVICE_DATA pDrvData;

pDrvData = device_get_softc (dev);

if (pDrvData->bDSPEnabled) {
     tp3780I_DisableDSP(dev, &pDrvData->rBDData);
     }

bus_generic_suspend (dev);
return 0;
}

static int mwave_resume (device_t dev)
{
pMWAVE_DEVICE_DATA pDrvData;
int i;

pDrvData = device_get_softc (dev);

for(i=0;i<16;i++) {
    pDrvData->IPCs[i].bIsEnabled=FALSE;
    pDrvData->IPCs[i].bIsHere=FALSE;
  }


smapi_set_DSP_power_state(FALSE);
smapi_set_DSP_power_state(TRUE);

if (!dsp3780I_EnableDSP (&pDrvData -> rBDData.rDspSettings, 
    s_ausThinkpadIrqToField, s_ausThinkpadDmaToField))
    PRINTK_ERROR ("mwave.c::mwave_resume, dsp7880I_EnableDSP failed\n");

tp3780I_EnableSRAM (&pDrvData -> rBDData);

bus_generic_resume (dev);
return 0;
}

static device_method_t mwave_methods[] =
    {
    DEVMETHOD(device_identify,	mwave_identify),
    DEVMETHOD(device_probe,	mwave_probe),
    DEVMETHOD(device_attach,	mwave_attach),
    DEVMETHOD(device_detach,	mwave_detach),
    DEVMETHOD(device_suspend,	mwave_suspend),
    DEVMETHOD(device_resume,	mwave_resume),
    {0, 0}
    };

static driver_t mwave_driver =
    {
    "mwave",
    mwave_methods,
    sizeof (mwave_softc_t)
    };

DRIVER_MODULE (mwave, isa, mwave_driver, mwave_devclass, 0, 0);
/* MODULE_VERSION (mwave, 1); */
