/*
 * smapi.c
 * Thinkpad SMAPI interface routines.
 *
 *  Written By: Mike Sullivan IBM Corporation
 *
 *  Copyright (C) 1999 IBM Corporation
 *
 * This program is free software; you can redistribute it and/or modify      
 * it under the terms of the GNU General Public License as published by      
 * the Free Software Foundation; either version 2 of the License, or         
 * (at your option) any later version.                                       
 *                                                                           
 * This program is distributed in the hope that it will be useful,           
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             
 * GNU General Public License for more details.                              
 *                                                                           
 * NO WARRANTY                                                               
 * THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR        
 * CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT      
 * LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,      
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is    
 * solely responsible for determining the appropriateness of using and       
 * distributing the Program and assumes all risks associated with its        
 * exercise of rights under this Agreement, including but not limited to     
 * the risks and costs of program errors, damage to or loss of data,         
 * programs or equipment, and unavailability or interruption of operations.  
 *                                                                           
 * DISCLAIMER OF LIABILITY                                                   
 * NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY   
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL        
 * DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND   
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    
 * USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED  
 * HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES             
 *                                                                           
 * You should have received a copy of the GNU General Public License         
 * along with this program; if not, write to the Free Software               
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *                                                                           
 * 
 *  10/23/2000 - Alpha Release 0.1.0
 *            First release to the public
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <machine/clock.h>

#define MW_TRACE
#include "smapi.h"
#include "mwavedd.h"

BOOLEAN smapi_request(unsigned short inBX, unsigned short inCX, 
    unsigned short inDI, unsigned short inSI,
    unsigned short *outAX, unsigned short *outBX, 
    unsigned short *outCX,
    unsigned short *outDX, unsigned short *outDI, 
    unsigned short *outSI);

static unsigned short   g_usSmapiPort=0;
int dspirq, dspio, uartirq, uartio;

BOOLEAN smapi_init()
{
  BOOLEAN bRC=FALSE;
  unsigned short usSmapiID=0;

  PRINTK_1(TRACE_SMAPI,"smapi::smapi_init entry\n");
  
  usSmapiID=rtcin(0x7C);
  usSmapiID |= ( rtcin(0x7D) << 8);
  PRINTK_2(TRACE_SMAPI,"smapi::smapi_init usSmapiID %x\n",usSmapiID);

  if (usSmapiID == 0x5349) {
    g_usSmapiPort=rtcin(0x7E);
    g_usSmapiPort |= (rtcin(0x7F) << 8);
    if (g_usSmapiPort == 0) {
      PRINTK_ERROR("smapi::smapi_init, ERROR unable to read from SMAPI port\n");
    } else {
      PRINTK_2(TRACE_SMAPI,"smapi::smapi_init, exit TRUE g_usSmapiPort %x\n",g_usSmapiPort);
      bRC=TRUE;
      //SmapiQuerySystemID();
    }
  } else {
    PRINTK_ERROR("smapi::smapi_init, ERROR invalid usSmapiID\n");
  }
  return bRC;
}


static BOOLEAN dummy_smapi_request(unsigned short inBX, unsigned short inCX, unsigned short inDI, unsigned short inSI,
                      unsigned short *outAX, unsigned short *outBX, unsigned short *outCX,
                      unsigned short *outDX, unsigned short *outDI, unsigned short *outSI) {

__asm__ __volatile__ (
	".type smapi_request,@function;"
	"smapi_request:;"
	"pushl %ebp;"
	"movl %esp, %ebp;"
	"pushl %ebx;"
	"pushl %ecx;"
	"pushl %edx;"
	"pushl %esi;"
	"pushl %edi;"
	"movw $0x5380, %ax;"
	"movw 8(%ebp), %bx;"
	"movw 12(%ebp), %cx;"
	"movw 16(%ebp), %di;"
	"movw 20(%ebp), %si;"
	"movw (g_usSmapiPort), %dx;"
	"pushl %ebp;"
	"outb  %al, %dx;"
	"outb  %al, $0x4F;"

	"popl  %ebp;"
	"orb  %ah, %ah;"
	"jnz  2f;"
	"pushl %edi;"
	"movl 24(%ebp), %edi;"
	"movw %ax, (%edi);"
        "movl 28(%ebp), %edi;"
        "movw %bx, (%edi);"
        "movl 32(%ebp), %edi;"
        "movw %cx, (%edi);"
        "movl 36(%ebp), %edi;"
        "movw %dx, (%edi);"
        "movl 44(%ebp), %edi;"
        "movw %si, (%edi);"
        "movl 40(%ebp), %edi;"
	"popl  %ebx;"
        "movw %bx, (%edi);"
	"movl $1, %eax;"
	"popl %edi;"
	"popl %esi;"
	"popl %edx;"
	"popl %ecx;"
	"popl %ebx;"
	"popl %ebp;"
	"ret;"

	"2:;"
	"xorl %eax, %eax;"
        "popl %edi;"
        "popl %esi;"
        "popl %edx;"
        "popl %ecx;"
        "popl %ebx;"
        "popl %ebp;"
        "ret;"

	);
return 1;
}


BOOLEAN smapi_query_DSP_cfg(SMAPI_DSP_SETTINGS *pSettings) {
  BOOLEAN bRC=FALSE;
  unsigned short usAX, usBX, usCX, usDX, usDI, usSI;
  unsigned short ausDspBases[]= { 0x0030, 0x4E30, 0x8E30, 0xCE30, 0x0130, 0x0350, 0x0070, 0x0DB0 };
  unsigned short ausUartBases[]= {0x03F8, 0x02F8, 0x03E8, 0x02E8};
  unsigned short numDspBases=8;
  unsigned short numUartBases=4;

  PRINTK_1(TRACE_SMAPI,"smapi::smapi_query_DSP_cfg entry\n");

  bRC=smapi_request(0x1802, 0x0000, 0, 0,
		   &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
  if (bRC == FALSE) {
    PRINTK_ERROR("smapi::smapi_query_DSP_cfg, get DSP Settings smapi_request FAILED\n");
  } else {
    PRINTK_1(TRACE_SMAPI,"smapi::smapi_query_DSP_cfg, smapi_request OK\n");
    pSettings->bDSPPresent=((usBX & 0x0100) != 0);
    pSettings->bDSPEnabled=((usCX & 0x0001) != 0);
    pSettings->usDspIRQ   = usSI & 0x00FF;
    pSettings->usDspDMA   = (usSI & 0xFF00) >> 8;
    if ((usDI & 0x00FF) < numDspBases) {
      pSettings->usDspBaseIO=ausDspBases[usDI & 0x00FF];
    } else {
      pSettings->usDspBaseIO=0;
    }
    
    PRINTK_6(TRACE_SMAPI,"smapi::smapi_query_DSP_cfg get DSP Settings bDSPPresent %x bDSPEnabled %x usDspIRQ %x usDspDMA %x usDspBaseIO %x\n",
	     pSettings->bDSPPresent,pSettings->bDSPEnabled,pSettings->usDspIRQ,pSettings->usDspDMA,pSettings->usDspBaseIO);

        /* check for illegal values */
        if ( pSettings->usDspBaseIO == 0 )
                printf("smapi::smapi_query_DSP_cfg: Worry: DSP base I/O address is 0\n");
        if ( pSettings->usDspIRQ == 0 )
                printf("smapi::smapi_query_DSP_cfg: Worry: DSP IRQ line is 0\n");

    bRC=smapi_request(0x1804, 0x0000, 0, 0, &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
    if (bRC == FALSE) {
      PRINTK_ERROR("smapi::smapi_query_DSP_cfg, get DSP modem settings FAILED\n");
    } else {
      pSettings->bModemEnabled=((usCX & 0x0001) != 0);
      pSettings->usUartIRQ=usSI & 0x000F;
      if (((usSI & 0xFF00) >> 8) < numUartBases) {
	pSettings->usUartBaseIO=ausUartBases[(usSI & 0xFF00) >> 8];
      } else {
	pSettings->usUartBaseIO=0;
      }
      PRINTK_4(TRACE_SMAPI,"smapi::smapi_query_DSP_cfg get DSP modem settings bModemEnabled %x usUartIRQ %x usUartBaseIO %x\n",
		pSettings->bModemEnabled,pSettings->usUartIRQ,pSettings->usUartBaseIO);
      PRINTK_ERROR("Mwave Modem, UART settings IRQ 0x%x   IO 0x%x\n",pSettings->usUartIRQ,pSettings->usUartBaseIO);

        /* check for illegal values */
        if ( pSettings->usUartBaseIO == 0 )
                printf("smapi::smapi_query_DSP_cfg: Worry: UART base I/O address is 0\n");
        if ( pSettings->usUartIRQ == 0 )
                printf("smapi::smapi_query_DSP_cfg: Worry: UART IRQ line is 0\n");

    }
  }
  PRINTK_2(TRACE_SMAPI,"smapi::smapi_query_DSP_cfg exit bRC %x\n",bRC);
  return bRC;
}

BOOLEAN smapi_set_DSP_cfg() {
  BOOLEAN bRC=FALSE;
  int i;
  unsigned short usAX, usBX, usCX, usDX, usDI, usSI;
  unsigned short ausDspBases[]= { 0x0030, 0x4E30, 0x8E30, 0xCE30, 0x0130, 0x0350, 0x0070, 0x0DB0 };
  unsigned short ausUartBases[]= {0x03F8, 0x02F8, 0x03E8, 0x02E8};
  unsigned short ausDspIrqs[]={ 5, 7, 10, 11, 15 };
  unsigned short ausUartIrqs[]={3, 4};
  
  unsigned short numDspBases=8;
  unsigned short numUartBases=4;
  unsigned short numDspIrqs=5;
  unsigned short numUartIrqs=2;
  unsigned short dspio_index=0, uartio_index=0;

  PRINTK_5(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg entry dspirq %x dspio %x uartirq %x uartio %x\n",dspirq,dspio,uartirq,uartio);

  if (dspio) {
    for(i=0; i<numDspBases; i++) {
      if (dspio == ausDspBases[i]) break;
    }
    if (i == numDspBases) {
      PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg ERROR invalid dspio address %x\n",dspio);
      return bRC;
    }
    dspio_index=i;
  }

  if (dspirq) {
    for(i=0; i<numDspIrqs; i++) {
      if (dspirq == ausDspIrqs[i]) break;
    }
    if (i == numDspIrqs) {
      PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg ERROR invalid dspirq %x\n",dspirq);
      return bRC;
    }
  }

  if (uartio) {
    for(i=0; i<numUartBases; i++) {
      if (uartio == ausUartBases[i]) break;
    }
    if (i == numUartBases) {
      PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg ERROR invalid uartio address %x\n",uartio);
      return bRC;
    }
    uartio_index=i;
  }


  if (uartirq) {
    for(i=0; i<numUartIrqs; i++) {
      if (uartirq == ausUartIrqs[i]) break;
    }
    if (i == numDspBases) {
      PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg ERROR invalid uartirq %x\n",uartirq);
      return bRC;
    }
  }

  if (uartirq || uartio) {
    bRC=smapi_request(0x1402, 0x0000, 0, 0,
		      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
    if (bRC) {
      if (usBX & 0x0100) { /* serial port A is present */
        if (usCX & 1) { /* serial port is enabled */
          if ( (usSI & 0xFF) == uartirq) {
	    PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR Serial Port A irq conflict usSI %x uartirq %x\n",usSI, uartirq);
#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	    PRINTK_1(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg Disabling conflicting Serial Port\n");
	    bRC=smapi_request(0x1403, 0x0100, 0, usSI,
			      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	    bRC=smapi_request(0x1402, 0x0000, 0, 0,
			      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif
	  } else {
	    if ( (usSI >> 8) == uartio_index) {
	      PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR Serial Port A IO conflict usSI %x uartio_index %x\n",usSI, uartio_index);

#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	      PRINTK_1(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg Disabling conflicting Serial Port\n");
	      bRC=smapi_request(0x1403, 0x0100, 0, usSI,
				&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	      bRC=smapi_request(0x1402, 0x0000, 0, 0,
				&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif
	    }
	  }
        }
      }
    }
    
    bRC=smapi_request(0x1404, 0x0000, 0, 0,
		      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
    if (bRC) {
      if (usBX & 0x0100) { /* serial port B is present */
        if (usCX & 1) { /* serial port is enabled */
          if ( (usSI & 0xFF) == uartirq) {
	    PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR Serial Port B irq conflict usSI %x uartirq %x\n",usSI, uartirq);

#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	    PRINTK_1(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg Disabling conflicting Serial Port\n");
	    bRC=smapi_request(0x1405, 0x0100, 0, usSI,
			      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	    bRC=smapi_request(0x1404, 0x0000, 0, 0,
			      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif

	  } else {
	    if ( (usSI >> 8) == uartio_index) {
	      PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR Serial Port B IO conflict usSI %x uartio_index %x\n",usSI, uartio_index);

#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	      PRINTK_1(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg Disabling conflicting Serial Port\n");
	      bRC=smapi_request(0x1405, 0x0100, 0, usSI,
				&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	      bRC=smapi_request(0x1404, 0x0000, 0, 0,
				&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif

            }
          }
        }
      }
    }

    /* Check IR port */
    bRC=smapi_request(0x1700, 0x0000, 0, 0, &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);

    bRC=smapi_request(0x1704, 0x0000, 0, 0,
		      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
    if (bRC) {
      if ( (usCX & 0xff) == uartirq) { /* serial port is enabled */
	PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR IR irq conflict usSI %x uartirq %x\n",usSI, uartirq);

#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	PRINTK_1(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg Disabling conflicting IR Port\n");
        bRC=smapi_request(0x1701, 0x0100, 0, 0,
			&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	bRC=smapi_request(0x1700, 0,0,0,&usAX,&usBX,&usCX,&usDX,&usDI,&usSI);
	bRC=smapi_request(0x1705, 0x01ff, 0, usSI,
			  &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	bRC=smapi_request(0x1704, 0x0000, 0, 0,
			  &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif

      } else {
	if ( (usSI & 0xff) == uartio_index) {
	  PRINTK_3(TRACE_SMAPI, "smapi::smapi_set_DSP_cfg ERROR IR IO conflict usSI %x uartio_index %x\n",usSI, uartio_index);

#ifdef MWAVE_FUTZ_WITH_OTHER_DEVICES
	  bRC=smapi_request(0x1701, 0x0100, 0, 0,
			&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	  bRC=smapi_request(0x1700, 0,0,0,&usAX,&usBX,&usCX,&usDX,&usDI,&usSI);
	  bRC=smapi_request(0x1705, 0x01ff, 0, usSI,
			    &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	  bRC=smapi_request(0x1704, 0x0000, 0, 0,
			    &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
#endif

	}
      }
    }
  }
  
  bRC=smapi_request(0x1802, 0x0000, 0, 0,
		    &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
  if (bRC) {
    if (dspio) {
      usDI=dspio_index;;
    } 
    if (dspirq) {
      usSI = (usSI & 0xff00) | dspirq;
    }
    
    bRC=smapi_request(0x1803, 0x0101, usDI, usSI, 
		      &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
    if (bRC) {
      bRC=smapi_request(0x1804, 0x0000, 0, 0,
			&usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
      if (bRC) {
	if (uartio) {
	  usSI = (usSI & 0x00ff) | (uartio_index<<8);
	}
	if (uartirq) {
	  usSI = (usSI & 0xff00) | uartirq;
	}
	bRC=smapi_request(0x1805, 0x0101, 0, usSI,
			  &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
	
      }
    }
  }    

  bRC=smapi_request(0x1802, 0x0000, 0, 0,
		    &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
  
  bRC=smapi_request(0x1804, 0x0000, 0, 0,
		    &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);
  
  PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_cfg exit bRC %x\n",bRC);
  return bRC;
}


BOOLEAN smapi_set_DSP_power_state(BOOLEAN bOn)
{
  BOOLEAN bRC=FALSE;
  unsigned short usAX, usBX, usCX, usDX, usDI, usSI;
  unsigned short usPowerFunction;

  PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_power_state entry bOn %x\n",bOn);

  usPowerFunction=(bOn) ? 1:0;

  bRC=smapi_request(0x4901, 0x0000, 0, usPowerFunction,
		       &usAX, &usBX, &usCX, &usDX, &usDI, &usSI);

  PRINTK_2(TRACE_SMAPI,"smapi::smapi_set_DSP_power_state exit bRC %x\n",bRC);
  return bRC;
}

#if 0
BOOLEAN SmapiQuerySystemID() {
  BOOLEAN bRC=FALSE;
  unsigned short usAX=0xffff, usBX=0xffff, usCX=0xffff, usDX=0xffff, usDI=0xffff, usSI=0xffff;

  printf("smapi::SmapiQUerySystemID entry\n");
  bRC=smapi_request(0x0000,0,0,0,&usAX,&usBX,&usCX,&usDX,&usDI, &usSI);

  if(bRC) {
    printf("AX=%x, BX=%x, CX=%x, DX=%x, DI=%x, SI=%x\n",usAX,usBX,usCX,usDX,usDI,usSI);
  } else {
    printf("smapi::SmapiQuerySystemID smapi_request error\n");
  }
  return bRC;
}
#endif
