/*
 * tp3780i.c
 * Board driver for 3780i based thinkpads.
 *
 *  Written By: Mike Sullivan IBM Corporation
 *
 *  Copyright (C) 1999 IBM Corporation
 *
 * This program is free software; you can redistribute it and/or modify      
 * it under the terms of the GNU General Public License as published by      
 * the Free Software Foundation; either version 2 of the License, or         
 * (at your option) any later version.                                       
 *                                                                           
 * This program is distributed in the hope that it will be useful,           
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             
 * GNU General Public License for more details.                              
 *                                                                           
 * NO WARRANTY                                                               
 * THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR        
 * CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT      
 * LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,      
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is    
 * solely responsible for determining the appropriateness of using and       
 * distributing the Program and assumes all risks associated with its        
 * exercise of rights under this Agreement, including but not limited to     
 * the risks and costs of program errors, damage to or loss of data,         
 * programs or equipment, and unavailability or interruption of operations.  
 *                                                                           
 * DISCLAIMER OF LIABILITY                                                   
 * NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY   
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL        
 * DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND   
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    
 * USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED  
 * HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES             
 *                                                                           
 * You should have received a copy of the GNU General Public License         
 * along with this program; if not, write to the Free Software               
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *                                                                           
 * 
 *  10/23/2000 - Alpha Release 0.1.0
 *            First release to the public
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>   
#include <sys/bus.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/rman.h>
#include <sys/resource.h>
#include <sys/ioccom.h>

#undef MW_TRACE

#include "smapi.h"
#include "mwavedd.h"
#include "tp3780i.h"
#include "3780i.h"
#include "mwavepub.h"

unsigned short s_ausThinkpadIrqToField[16] = {0xFFFF,0xFFFF,0xFFFF,0x0001,0x0002,0x0003,0xFFFF,0x0004,
						     0xFFFF,0xFFFF,0x0005,0x0006,0xFFFF,0xFFFF,0xFFFF,0x0007};
unsigned short s_ausThinkpadDmaToField[8]  = {0x0001,0x0002,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0x0003,0x0004};
static unsigned short s_numIrqs=16, s_numDmas=8;

static void DspInterrupt(void*);
static void UartInterrupt(void*);

extern int dspirq, dspio, uartirq, uartio;

int mwave_Initialize(THINKPAD_BD_DATA *pBDData)
{
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::Initialize entry pBDData %p\n",pBDData);
    
  pBDData->bDSPEnabled=FALSE;
  pSettings->bInterruptClaimed=FALSE;

  if (smapi_init() == FALSE) {
    retval= EIO;
    
    PRINTK_ERROR("tp3780i::Initialize SMAPI is not available on this machine\n");
  } else {
    if (dspirq || dspio || uartirq || uartio ) {
      if (smapi_set_DSP_cfg() == FALSE) {
        retval= EIO;
      } 
       else {
	return EAGAIN;  
      } 
    }
  }

  
  PRINTK_2(TRACE_TP3780I,"tp3780i::Initialize exit retval %x\n",retval);

  return retval;
}

#if 0
static int Cleanup(THINKPAD_BD_DATA *pBDData) 
{
  int retval=0;

  
  PRINTK_2(TRACE_TP3780I,"tp3780i::Cleanup entry and exit pBDData %p\n",pBDData);
  return retval;
}
#endif

int mwave_CalcResources(THINKPAD_BD_DATA *pBDData)
{
  int retval=0;
  SMAPI_DSP_SETTINGS rSmapiInfo;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;

  
  PRINTK_2(TRACE_TP3780I,"tp3780i::CalcResources entry pBDData %p\n",pBDData);

  if (smapi_query_DSP_cfg(&rSmapiInfo) == TRUE) {
    pSettings->bDSPEnabled=(rSmapiInfo.bDSPEnabled && rSmapiInfo.bDSPPresent);
    pSettings->bModemEnabled=rSmapiInfo.bModemEnabled;
    pSettings->usDspIrq=rSmapiInfo.usDspIRQ;
    pSettings->usDspDma=rSmapiInfo.usDspDMA;
    pSettings->usDspBaseIO=rSmapiInfo.usDspBaseIO;
    pSettings->usUartIrq=rSmapiInfo.usUartIRQ;
    pSettings->usUartBaseIO=rSmapiInfo.usUartBaseIO;
    
    pSettings->uDStoreSize=TP_ABILITIES_DATA_SIZE;
    pSettings->uIStoreSize=TP_ABILITIES_INST_SIZE;
    pSettings->uIps=TP_ABILITIES_INTS_PER_SEC;
    
    if (pSettings->bDSPEnabled && pSettings->bModemEnabled && pSettings->usDspIrq == pSettings->usUartIrq) {
      pBDData->bShareDspIrq=pBDData->bShareUartIrq=1;
    } else {
      pBDData->bShareDspIrq=pBDData->bShareUartIrq=0;
    }
  } else {
    retval= EIO;
  }
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::CalcResources exit retval %x\n",retval);
  return retval;
}


int mwave_ClaimResources(THINKPAD_BD_DATA *pBDData)
{
  int retval=0;

  PRINTK_2(TRACE_TP3780I,"tp3780i::ClaimResources entry pBDData %p\n",pBDData);

#if 0
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;

  if (!check_region(pSettings->usDspBaseIO,16)) {
    
    PRINTK_1(TRACE_TP3780I,"tp3780i::ClaimResources check region returns 0\n");
    request_region(pSettings->usDspBaseIO,16,"mwave_3780i");
  } else {
    retval= EIO;
    PRINTK_ERROR("tp3780i::ClaimResources check region Error\n");
  }
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::ClaimResources exit retval %x\n",retval);
#endif

  return retval;
}

int mwave_ReleaseResources (device_t dev, THINKPAD_BD_DATA *pBDData)
{
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;

  
  PRINTK_2(TRACE_TP3780I,"tp3780i::mwave_ReleaseResources entry pBDData %p\n",pBDData);

  if (pSettings->bInterruptClaimed) {
      if ((retval=bus_teardown_intr (dev, pBDData -> irq_res, pBDData -> irq_id)))
            {
            printf ("bus_teardown_intr: - %d\n", retval);
            return retval;
            }
      if ((retval = bus_release_resource (dev, SYS_RES_IRQ, 0, pBDData ->irq_res)))
            {
            printf ("bus_release_resource: - %d\n", retval);
            return retval;
            }

      pSettings->bInterruptClaimed=FALSE;
  }

  if ((retval = bus_release_resource (dev, SYS_RES_IOPORT, 0,
	pBDData -> port_res)))
	{
	printf ("mwavem: bus_release_resource DSP port: %d\n", retval);
	return retval;
	}
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::mwave_ReleaseResources exit retval %x\n",retval);

  return retval;
}


int tp3780I_EnableDSP(device_t dev, THINKPAD_BD_DATA *pBDData)
{
  int retval= EIO;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  BOOLEAN bDSPPoweredUp=FALSE, bDSPEnabled=FALSE, bInterruptAllocated=FALSE;
  int rid = 0;

  
  PRINTK_2(TRACE_TP3780I,"tp3780i::tp3780I_EnableDSP entry pBDData %p\n",pBDData);

  if (pBDData->bDSPEnabled) {
    PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, pBDData->bDSPEnabled %x already!!\n",pBDData->bDSPEnabled);
    goto cleanup;
  }

  if (!pSettings->bDSPEnabled) {
    PRINTK_ERROR("tp3780::tp3780I_EnableDSP, pSettings->bDSPEnabled %x failed\n",pSettings->bDSPEnabled);
    goto cleanup;
  }
  
  if ((pSettings->usDspIrq >= s_numIrqs) || (pSettings->usDspDma >= s_numDmas) ||
      (s_ausThinkpadIrqToField[pSettings->usDspIrq] == 0xFFFF) || 
      (s_ausThinkpadDmaToField[pSettings->usDspDma] == 0xFFFF)) {
    PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, invalid irq mapping %x\n",pSettings->usDspIrq);
    goto cleanup;
  }

  if (((pSettings->usDspBaseIO & 0xF00F) != 0) || (pSettings->usDspBaseIO & 0x0FF0) == 0) {
    PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, invalid DspBaseIO %x\n",pSettings->usDspBaseIO);
    goto cleanup;
  }

  if (pSettings->bModemEnabled) {
    if (pSettings->usUartIrq >= s_numIrqs || s_ausThinkpadIrqToField[pSettings->usUartIrq] == 0xFFFF) {
      PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, invalid Uart irq mapping %x\n",pSettings->usUartIrq);
      goto cleanup;
    }
    switch(pSettings->usUartBaseIO) {
    case 0x03F8:
    case 0x02F8:
    case 0x03E8:
    case 0x02E8:
      break;

    default:
      PRINTK_ERROR("tp3780i::EnabledDSP invalid usUartBaseIO %x\n",pSettings->usUartBaseIO);
      goto cleanup;
    }
  }

  pSettings->bDspIrqActiveLow=pSettings->bDspIrqPulse=TRUE;
  pSettings->bUartIrqActiveLow=pSettings->bUartIrqPulse=TRUE;

  if (pBDData->bShareDspIrq) {
    pSettings->bDspIrqActiveLow=FALSE;
  }
  if (pBDData->bShareUartIrq) {
    pSettings->bUartIrqActiveLow=FALSE;
  }

  pSettings->usNumTransfers=TP_CFG_NumTransfers;
  pSettings->usReRequest=TP_CFG_RerequestTimer;
  pSettings->bEnableMEMCS16=TP_CFG_MEMCS16;
  pSettings->usIsaMemCmdWidth=TP_CFG_IsaMemCmdWidth;
  pSettings->bGateIOCHRDY=TP_CFG_GateIOCHRDY;
  pSettings->bEnablePwrMgmt=TP_CFG_EnablePwrMgmt;
  pSettings->usHBusTimerLoadValue=TP_CFG_HBusTimerValue;
  pSettings->bDisableLBusTimeout=TP_CFG_DisableLBusTimeout;
  pSettings->usN_Divisor=TP_CFG_N_Divisor;
  pSettings->usM_Multiplier=TP_CFG_M_Multiplier;
  pSettings->bPllBypass=TP_CFG_PllBypass;
  pSettings->usChipletEnable=TP_CFG_ChipletEnable;

  /* Test for Uart IRQ conflicts */

  /* I'm sure this isn't how it's supposed to be done */
  if ((retval = bus_set_resource (dev, SYS_RES_IRQ, rid, pSettings->usUartIrq,
	1)))
	{
	printf ("mwave: bus_set_resource failed\n");
	return retval;
	}
   pBDData -> irq_res = bus_alloc_resource (dev, SYS_RES_IRQ, &rid, pSettings->usUartIrq,
	 pSettings->usUartIrq, 1, RF_SHAREABLE|RF_ACTIVE);
   if (!pBDData -> irq_res)
	{
	printf ("mwave: can't allocate IRQ resource\n");
	return EIO;
	}
  if (bus_setup_intr (dev, pBDData -> irq_res, INTR_TYPE_TTY,
	&UartInterrupt, 0, &pBDData -> irq_id))
	{ 
	printf ("mwave: can't setup uart interrupt");
	bus_release_resource (dev, SYS_RES_IRQ, rid, pBDData -> irq_res);
	return EIO;
	}
    
  if ((retval = bus_teardown_intr (dev, pBDData -> irq_res, pBDData -> irq_id)))
	{
	printf ("bus_teardown_intr: - %d\n", retval);
	return retval;
	}
  if ((retval = bus_release_resource (dev, SYS_RES_IRQ, rid, pBDData -> irq_res)))
	{
	printf ("bus_release_resource: - %d\n", retval);
	return retval;
	}

    bInterruptAllocated=TRUE;
    pSettings->bInterruptClaimed=TRUE;

   rid = 0;

   if ((retval = bus_set_resource (dev, SYS_RES_IOPORT, rid,
        pSettings->usDspBaseIO, 16)) != 0)
	{
	printf ("mwave: can't set DSP port resource\n");
	return retval;
	}
   pBDData -> port_res = bus_alloc_resource (dev, SYS_RES_IOPORT, &rid,
 	0ul, ~0ul, 1, RF_ACTIVE);
   if (pBDData -> port_res == 0)
	{
	printf ("mwavem: bus_alloc_resource dsp failed\n");
	return EIO;
	}

   rid = 0;
   if ((retval = bus_set_resource (dev, SYS_RES_IRQ, rid, pSettings -> usDspIrq,
	1)))
	{
	printf ("mwave: bus_set_resource dsp failed\n");
	bus_release_resource (dev, SYS_RES_IOPORT, 0, pBDData -> port_res);
	return retval;
	}
   pBDData -> irq_res = bus_alloc_resource (dev, SYS_RES_IRQ, &rid, 
	pSettings -> usDspIrq, pSettings -> usDspIrq, 1, RF_SHAREABLE|RF_ACTIVE);
   if (!pBDData -> irq_res)
        {
        printf ("mwave: can't allocate dsp IRQ resource\n");
	bus_release_resource (dev, SYS_RES_IOPORT, 0, pBDData -> port_res);
        return EIO;
        }

  pBDData -> irq_id = 0;
  if (bus_setup_intr (dev, pBDData -> irq_res, INTR_TYPE_TTY,
        &DspInterrupt, (void*) pBDData, &pBDData -> irq_id))
        {
        printf ("mwave: can't setup dsp interrupt");
        bus_release_resource (dev, SYS_RES_IRQ, rid, pBDData -> irq_res);
	bus_release_resource (dev, SYS_RES_IOPORT, 0, pBDData -> port_res);
        return EIO;
	}
  smapi_set_DSP_power_state(FALSE);
  if (!smapi_set_DSP_power_state(TRUE)) {
    PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, smapi_set_DSP_power_state(1) fails\n");
    goto cleanup;
  } else {
    bDSPPoweredUp=TRUE;
  }

  if (!dsp3780I_EnableDSP(pSettings, s_ausThinkpadIrqToField, s_ausThinkpadDmaToField)) {
    PRINTK_ERROR("tp3780i::tp3780I_EnableDSP, dsp7880I_EnableDSP failed\n");
    goto cleanup;
  } else {
    bDSPEnabled=TRUE;
  }
  
  tp3780I_EnableSRAM(pBDData);

  retval=0;
  pBDData->bDSPEnabled=TRUE;
  PRINTK_2(TRACE_TP3780I,"tp3780i::tp3780I_EnableDSP exit retval %x\n",retval);
  return retval;
  
 cleanup:   
  PRINTK_1(TRACE_TP3780I,"tp3780i::tp3780I_EnableDSP error, performing cleanup\n");
  if (bDSPEnabled) dsp3780I_DisableDSP(pSettings);
  if (bDSPPoweredUp) smapi_set_DSP_power_state(FALSE);
  if (bInterruptAllocated) { 
        pSettings->bInterruptClaimed=FALSE;
        }
  return EIO;
  
}


int tp3780I_DisableDSP(device_t dev, THINKPAD_BD_DATA *pBDData) {
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::tp3780I_DisableDSP entry pBDData %p\n",pBDData);

  if (pBDData->bDSPEnabled) {
    dsp3780I_DisableDSP(&pBDData->rDspSettings);
    if (pSettings->bInterruptClaimed) {
      if ((retval = bus_teardown_intr (dev, pBDData -> irq_res, pBDData -> irq_id)))
	{
	printf ("mwave: can't teardown irq - %d\n", retval);
	return retval;
	}
      if ((retval = bus_release_resource (dev, SYS_RES_IRQ, 0, pBDData -> irq_res)))
	{
	printf ("mwave: can't release irq resource - %d", retval);
	return retval;
	}
      pSettings->bInterruptClaimed=FALSE;
    }
    smapi_set_DSP_power_state(FALSE);
    pBDData->bDSPEnabled=FALSE;
  }

  PRINTK_2(TRACE_TP3780I,"tp3780i::tp3780I_DisableDSP exit retval %x\n",retval);

  return retval;
}

int tp3780I_ResetDSP(THINKPAD_BD_DATA *pBDData) {
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;

  PRINTK_2(TRACE_TP3780I,"tp3780i::ResetDSP entry pBDData %p\n",pBDData);

  if (dsp3780I_Reset(pSettings) == TRUE) {
    tp3780I_EnableSRAM(pBDData);
  } else {
    retval= EIO;
  }
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::ResetDSP exit retval %x\n",retval);
  return retval;
}


int tp3780I_StartDSP(THINKPAD_BD_DATA *pBDData) {
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;

  PRINTK_2(TRACE_TP3780I,"tp3780i::StartDSP entry pBDData %p\n",pBDData);

  if (dsp3780I_Run(pSettings) == TRUE) {
    // @BUG @TBD tp3780I_EnableSRAM(pBDData);
  } else {
    retval= EIO;
  }
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::StartDSP exit retval %x\n",retval);
  return retval;
}

int tp3780I_QueryAbilities(THINKPAD_BD_DATA *pBDData, MW_ABILITIES *pAbilities) {
  int retval=0;

  PRINTK_2(TRACE_TP3780I,"tp3780i::QueryAbilities entry pBDData %p\n",pBDData);

  /* fill out standard constant fields */
  pAbilities->instr_per_sec=pBDData->rDspSettings.uIps;
  pAbilities->data_size=pBDData->rDspSettings.uDStoreSize;
  pAbilities->inst_size=pBDData->rDspSettings.uIStoreSize;
  pAbilities->bus_dma_bw=pBDData->rDspSettings.uDmaBandwidth;

  /* fill out dynamically determined fields */
  pAbilities->component_list[0] = 0x00010000 | MW_ADC_MASK;
  pAbilities->component_list[1] = 0x00010000 | MW_ACI_MASK;
  pAbilities->component_list[2] = 0x00010000 | MW_AIC1_MASK;
  pAbilities->component_list[3] = 0x00010000 | MW_AIC2_MASK;
  pAbilities->component_list[4] = 0x00010000 | MW_CDDAC_MASK;
  pAbilities->component_list[5] = 0x00010000 | MW_MIDI_MASK;
  pAbilities->component_list[6] = 0x00010000 | MW_UART_MASK;
  pAbilities->component_count=7;

  /* Fill out Mwave OS and BIOS task names */

  memcpy(pAbilities->mwave_os_name ,TP_ABILITIES_MWAVEOS_NAME,sizeof(TP_ABILITIES_MWAVEOS_NAME));
  memcpy(pAbilities->bios_task_name,TP_ABILITIES_BIOSTASK_NAME,sizeof(TP_ABILITIES_BIOSTASK_NAME));

  PRINTK_1(TRACE_TP3780I,"tp3780i::QueryAbilities exit retval=SUCCESSFUL\n");
  return retval;
}

int tp3780I_ReadWriteDspDStore(THINKPAD_BD_DATA *pBDData, unsigned int uOpcode, void *pvBuffer,
		       unsigned int uCount, unsigned long ulDSPAddr)
{
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  unsigned short usDspBaseIO=pSettings->usDspBaseIO;
  BOOLEAN bRC=0;
  
  PRINTK_6(TRACE_TP3780I,"tp3780i::ReadWriteDspDStore entry pBDData %p, uOpcode %x, pvBuffer %p, uCount %x, ulDSPAddr %lx\n",
	    pBDData,uOpcode,pvBuffer,uCount,ulDSPAddr);

  if (pBDData->bDSPEnabled) {
    switch(uOpcode) {
    case IOCTL_MW_READ_DATA:
      bRC=dsp3780I_ReadDStore(usDspBaseIO,pvBuffer,uCount,ulDSPAddr);
      break;
      
    case IOCTL_MW_READCLEAR_DATA:
      bRC=dsp3780I_ReadAndClearDStore(usDspBaseIO,pvBuffer,uCount,ulDSPAddr);
      break;
      
    case IOCTL_MW_WRITE_DATA:
      bRC=dsp3780I_WriteDStore(usDspBaseIO,pvBuffer,uCount,ulDSPAddr);
      break;
    }
  }

  retval=(bRC) ? 0 :  EIO;
  PRINTK_2(TRACE_TP3780I,"tp3780i::ReadWriteDspDStore exit retval %x\n",retval);
  return retval;
}


int tp3780I_ReadWriteDspIStore(THINKPAD_BD_DATA *pBDData, unsigned int uOpcode, void *pvBuffer,
		       unsigned int uCount, unsigned long ulDSPAddr)
{
  int retval=0;
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  unsigned short usDspBaseIO=pSettings->usDspBaseIO;
  BOOLEAN bRC=0;
  
  PRINTK_6(TRACE_TP3780I,"tp3780i::ReadWriteDspIStore entry pBDData %p, uOpcode %x, pvBuffer %p, uCount %x, ulDSPAddr %lx\n",
	    pBDData,uOpcode,pvBuffer,uCount,ulDSPAddr);
  
  if (pBDData->bDSPEnabled) {
    switch(uOpcode) {
    case IOCTL_MW_READ_INST:
      bRC=dsp3780I_ReadIStore(usDspBaseIO,pvBuffer,uCount,ulDSPAddr);
      break;
      
    case IOCTL_MW_WRITE_INST:
      bRC=dsp3780I_WriteIStore(usDspBaseIO,pvBuffer,uCount,ulDSPAddr);
      break;
    }
  }

  retval=(bRC) ? 0 :  EIO;
  
  PRINTK_2(TRACE_TP3780I,"tp3780i::ReadWriteDspIStore exit retval %x\n",retval);
  return retval;
}
    

static void UartInterrupt(void *dev_id)
{
  PRINTK_2(TRACE_TP3780I,"tp3780i::UartInterrupt entry dev_id %x\n", (int)dev_id);
}

static void DspInterrupt(void* dev_data)
{
  pMWAVE_DEVICE_DATA pDrvData;
  DSP_3780I_CONFIG_SETTINGS *pSettings;
  unsigned short usDspBaseIO, usIPCSource = 0, usIsolationMask, usPCNum;

  pDrvData = (pMWAVE_DEVICE_DATA) dev_data; /* XXX: should be pBDData! */

  pSettings=&pDrvData->rBDData.rDspSettings;
  usDspBaseIO=pSettings->usDspBaseIO;

  PRINTK_1(TRACE_TP3780I,"tp3780i::DspInterrupt entry \n");

  if (dsp3780I_GetIPCSource(usDspBaseIO,&usIPCSource)) {
    PRINTK_2(TRACE_TP3780I,"tp3780i::DspInterrupt, return from dsp3780i_GetIPCSource, usIPCSource %x\n",usIPCSource);
      usIsolationMask=1;
      for(usPCNum=1; usPCNum <= 16; usPCNum++) {
        if (usIPCSource & usIsolationMask) {
  	  usIPCSource &= ~usIsolationMask;
	  PRINTK_3(TRACE_TP3780I,"tp3780i::DspInterrupt usPCNum %x usIPCSource %x\n",usPCNum,usIPCSource);
	  if ( pDrvData->IPCs[usPCNum-1].usIntCount == 0) {
	    pDrvData->IPCs[usPCNum-1].usIntCount=1;
	  }
	  PRINTK_2(TRACE_TP3780I,"tp3780i::DspInterrupt usIntCount %x\n",pDrvData->IPCs[usPCNum-1].usIntCount);
	  if (pDrvData->IPCs[usPCNum-1].bIsEnabled == TRUE) {
	    PRINTK_2(TRACE_TP3780I,"tp3780i::DspInterrupt, waking up usPCNum %x\n",usPCNum-1);
	    wakeup (&pDrvData->IPCs[usPCNum-1].ipc_wait_queue);
	  } else {
	    PRINTK_2(TRACE_TP3780I,"tp3780i::DspInterrupt, no one waiting for IPC %x\n",usPCNum-1);
	  }
        }
        if (usIPCSource == 0) break;
        /* try next IPC */
        usIsolationMask=usIsolationMask << 1;
      }
  } else {
    PRINTK_1(TRACE_TP3780I,"tp3780i::DspInterrupt, return false from dsp3780i_GetIPCSource\n");
  }
  PRINTK_1(TRACE_TP3780I,"tp3780i::DspInterrupt exit\n");
}



void tp3780I_EnableSRAM(THINKPAD_BD_DATA *pBDData)
{
  DSP_3780I_CONFIG_SETTINGS *pSettings=&pBDData->rDspSettings;
  unsigned short usDspBaseIO=pSettings->usDspBaseIO;
  DSP_GPIO_OUTPUT_DATA_15_8    rGpioOutputData;
  DSP_GPIO_DRIVER_ENABLE_15_8  rGpioDriverEnable;
  DSP_GPIO_MODE_15_8           rGpioMode;

  PRINTK_1(TRACE_TP3780I,"tp3780i::tp3780I_EnableSRAM, entry\n");

  MKWORD(rGpioMode)=ReadMsaCfg(DSP_GpioModeControl_15_8);
  rGpioMode.GpioMode10=0;
  WriteMsaCfg(DSP_GpioModeControl_15_8,MKWORD(rGpioMode));

  MKWORD(rGpioDriverEnable)=0;
  rGpioDriverEnable.Enable10=TRUE;
  rGpioDriverEnable.Mask10=TRUE;
  WriteMsaCfg(DSP_GpioDriverEnable_15_8,MKWORD(rGpioDriverEnable));

  MKWORD(rGpioOutputData)=0;
  rGpioOutputData.Latch10=0;
  rGpioOutputData.Mask10=TRUE;
  WriteMsaCfg(DSP_GpioOutputData_15_8,MKWORD(rGpioOutputData));

  PRINTK_1(TRACE_TP3780I,"tp3780i::tp3780I_EnableSRAM exit\n");
}
